"""logiskip's base code for loads"""

import logging
import sys
from typing import Any, Optional, Sequence, Union

from semantic_version import SimpleSpec, Version
from sqlalchemy import Table, create_engine, insert, select
from sqlalchemy.engine import Connection
from sqlalchemy.engine.base import Engine
from sqlalchemy.ext.automap import automap_base

if sys.version_info >= (3, 9):
    from importlib import metadata
else:
    import importlib_metadata as metadata

from .util import add_default_driver

_logger = logging.getLogger("logiskip")


class LoadRegistry:
    """Registry object that collects and finds available loads"""

    _loads: dict[str, dict[str, "BaseLoad"]]

    @staticmethod
    def import_known_loads() -> None:
        for ep in metadata.entry_points().get("logiskip.load", []):
            __import__(ep.module)
            _logger.debug("Found load %s", ep.module)

    def __init__(self):
        self._loads = {}

    def register(self, name: str, version_constraint: str, load_class: "BaseLoad") -> None:
        """Register a named load for given version constraints"""
        self._loads.setdefault(name, {})[version_constraint] = load_class
        _logger.debug("Registered load %s for versions %s", name, version_constraint)

    def find(self, name: str, version: Union[Version, str]) -> Optional["BaseLoad"]:
        """Find a load matching the given name and version"""
        if isinstance(version, str):
            version = Version(version)

        for constraint, load_class in self._loads.get(name, {}).items():
            spec = SimpleSpec(constraint)
            if spec.match(version):
                _logger.debug(
                    "Found load %s for versions %s, matching %s", name, constraint, version
                )
                return load_class

        _logger.error("No load %s found for version %s", name, version)
        return None


class BaseLoad:
    """Base class for logiskip load definitions"""

    @property
    def src_dialect(self) -> str:
        """Dialect name of source"""
        return self.src_engine.dialect.name

    @property
    def dest_dialect(self) -> str:
        """Dialect name of destination"""
        return self.dest_engine.dialect.name

    def __init_subclass__(
        cls, /, name: Optional[str] = None, version_constraint: str = "*", **kwargs
    ):
        """Register a load subclass so it can be found by name and version constraint"""
        super().__init_subclass__(**kwargs)

        if name is None:
            # Guess load name from class module
            if cls.__module__.startswith("logiskip.loads."):
                name = cls.__module__.split(".")[2]
            else:
                raise TypeError("No load name passed and load not in logiskip.loads namespace.")

        # Store information on class for later use
        cls.name = name
        cls.version_constraint = version_constraint

        load_registry.register(name, version_constraint, cls)

    def __init__(self, src: Union[Engine, str], dest: Union[Engine, str]):
        if isinstance(src, str):
            src = add_default_driver(src)
            self.src_engine = create_engine(src)
        else:
            self.src_engine = src
        self.src_base = automap_base()
        _logger.info("Discovering DDL for source (%s)", self.src_dialect)
        self.src_base.prepare(self.src_engine, reflect=True)

        if isinstance(dest, str):
            dest = add_default_driver(dest)
            self.dest_engine = create_engine(dest)
        else:
            self.dest_engine = dest
        self.dest_base = automap_base()
        _logger.info("Discovering DDL for destination (%s)", self.dest_dialect)
        self.dest_base.prepare(self.dest_engine, reflect=True)

    def get_dest_table_name(self, src_table_name: str) -> str:
        """Determine destination table name"""
        table_map = getattr(
            self,
            f"{self.src_dialect}_{self.dest_dialect}_tables",
            getattr(self, "default_tables", {}),
        )
        return table_map.get(src_table_name, src_table_name)

    def get_dest_field(self, src_table: Table, src_field: str) -> str:
        """Determine the destination field name"""
        field_map = getattr(
            self,
            f"{self.src_dialect}_{self.dest_dialect}_fields_{src_table.name}",
            getattr(self, f"default_fields_{src_table.name}", {}),
        )
        return field_map.get(src_field, src_field)

    def convert_table(self, src_table: Table) -> None:
        """Convert one table from source to destination"""
        _logger.info("Converting table %s", src_table.name)

        # Determine destination table
        dest_table_name = self.get_dest_table_name(src_table.name)
        if dest_table_name is None:
            # If dest_table_name is explicitly None, skip the table
            _logger.info("Skipping table %s (explicitly disabled)", src_table.name)
            return None
        _logger.debug("Table %s mapped to %s", src_table.name, dest_table_name)
        dest_table = self.dest_base.metadata.tables[dest_table_name]

        # Look for an explicit conversion method; resort to default implementation
        convert_meth = getattr(
            self,
            f"{self.src_dialect}_{self.dest_dialect}_table_{src_table.name}",
            getattr(self, f"default_table_{src_table.name}", self._convert_table_default),
        )
        return convert_meth(src_table, dest_table)

    def _convert_table_default(self, src_table: Table, dest_table: Table) -> None:
        # Extract all source rows
        src_stmt = src_table.select()
        dest_rows = []
        with self.src_engine.connect() as src_conn:
            for src_row in src_conn.execute(src_stmt):
                # Convert to destination row
                dest_row = self.convert_row(src_table, src_row._asdict())
                dest_rows.append(dest_row)

        if dest_rows:
            return dest_table.insert().values(dest_rows)
        return None

    def convert_row(self, src_table: Table, src_dict: dict[str, Any]) -> dict[str, Any]:
        """Convert a single row from a single table"""
        _logger.debug("Converting row %s", str(src_dict.values()))

        # Look for an explicit conversion method; resort to default implementation
        convert_meth = getattr(
            self,
            f"{self.src_dialect}_{self.dest_dialect}_row_{src_table.name}",
            getattr(self, f"default_row_{src_table.name}", self._convert_row_default),
        )
        return convert_meth(src_table, src_dict)

    def _convert_row_default(self, src_table: Table, src_dict: dict[str, Any]) -> dict[str, Any]:
        dest_dict = {}
        # Convert every single field
        for src_field, src_value in src_dict.items():
            # Determine destination field name
            dest_field = self.get_dest_field(src_table, src_field)
            if dest_field is None:
                # Skip field if explicitly set to None
                continue

            # Look for an explicit conversion method; resort to default implementation
            convert_meth = getattr(
                self,
                f"{self.src_dialect}_{self.dest_dialect}_field_{src_table.name}__{src_field}",
                getattr(
                    self,
                    f"default_field_{src_table.name}__{src_field}",
                    self._convert_field_default,
                ),
            )
            dest_value = convert_meth(src_value)

            dest_dict[dest_field] = dest_value

        return dest_dict

    def _convert_field_default(self, src_value: Any) -> Any:
        return src_value

    def migrate(self, commit: bool = True) -> None:
        """Run the migration defined by this load"""
        _logger.info("Migrating %s from %s to %s", self.name, self.src_dialect, self.dest_dialect)
        # Handle all known tables in order
        with self.dest_engine.connect() as dest_conn:
            with dest_conn.begin() as transaction:
                for src_table in self.src_base.metadata.tables.values():
                    dest_stmt = self.convert_table(src_table)
                    if dest_stmt is None:
                        # Skip if result is empty
                        continue
                    dest_conn.execute(dest_stmt)

                # Roll-back if commit is not desired
                if commit:
                    _logger.info(
                        "Migrated %s from %s to %s", self.name, self.src_dialect, self.dest_dialect
                    )
                else:
                    transaction.rollback()
                    _logger.warning("Transaction rolled back (explicitly requested)")


load_registry = LoadRegistry()
LoadRegistry.import_known_loads()



"""logiskip load definition for Roundcube"""

from logiskip.load import BaseLoad


class RoundcubeLoad(BaseLoad, name="roundcube", version_constraint=">=1.4,<1.6"):
    # The "system" table is pre-filled with the schema version
    default_tables = {"system": None}

    # The "dictionary" table has an "id" field in MySQL
    # Note from schema: redundant, for compat. with Galera Cluster
    # No other dialects have the field, so we can shorten it as default
    default_fields_dictionary = {"id": None}



"""logiskip's command-line interface"""

import logging

import click
import click_logging

from .load import load_registry

logging.basicConfig(format="%(asctime)s - %(levelname)s - %(message)s")
_logger = logging.getLogger("logiskip")


@click.command()
@click.option("--source", prompt="Source URI", help="URI of source database")
@click.option("--destination", prompt="Destination URI", help="URI of destination database")
@click.option(
    "--load-name", prompt="Load name", help="Name of load plugin for migrated application"
)
@click.option(
    "--load-version", prompt="Load version", help="Version of migrated application/schema"
)
@click.option(
    "--dry-run", help="Roll back transaction instead of commiting", is_flag=True, default=False
)
@click_logging.simple_verbosity_option(_logger)
def logiskip(
    source: str, destination: str, load_name: str, load_version: str, dry_run: bool
) -> None:
    """Main executable for logiskip"""
    load_class = load_registry.find(load_name, load_version)
    load = load_class(source, destination)
    load.migrate(commit=not dry_run)



"""Utility code for logiskip"""

default_drivers = {
    "mysql": "pymysql",
    "postgresql": "pg8000",
    "sqlite": "pysqlite",
    "oracle": "cx_oracle",
    "mssql": "pymssql",
}


def add_default_driver(uri: str) -> str:
    """Add the preferred dialect to an SQLAlchemy URI"""
    parts = uri.split("://", 1)
    if "+" in parts[0]:
        return uri
    else:
        driver = default_drivers.get(parts[0], None)
        if driver:
            parts[0] += f"+{driver}"
    return "://".join(parts)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['logiskip', 'logiskip.loads.roundcube']

package_data = \
{'': ['*']}

install_requires = \
['SQLAlchemy>=1.4.5,<2.0.0',
 'click-logging>=1.0.1,<2.0.0',
 'click>=7.1.2,<8.0.0',
 'semantic-version>=2.8.5,<3.0.0']

extras_require = \
{'mysql': ['PyMySQL>=1.0.2,<2.0.0'], 'postgres': ['pg8000>=1.19.1,<2.0.0']}

entry_points = \
{'console_scripts': ['logiskip = logiskip.cli:logiskip'],
 'logiskip.load': ['roundcube = logiskip.loads.roundcube']}

setup_kwargs = {
    'name': 'logiskip',
    'version': '0.1.0',
    'description': 'Logical database migration between RDBMSs',
    'long_description': 'logiskip - Logical database migration between RDBMSs \n====================================================\n\nSynopsis\n--------\n\nlogiskip is a command-line tool written in Python (using `SQLAlchemy`_ that cam\nmigrate application data between different RDMSs, e.g. between MySQL and PostgreSQL.\n\nIt is modular, with the possibility to define migration logic for applications in\nseparate classes (so-called "loads").\n\nlogiskip can be used for simple tasks like mere copying of tables, but also for\nmore complex tasks (e.g. converting images to another format, converting complex\ntypes, etc.).\n\nInstallation\n------------\n\nlogiskip can be installed using `pip`. In doing so, the needed database engines\ncan be passed as extras. To install logiskip with the ability to convert between\nMySQL and PostgreSQL::\n\n  pip3 install \'logiskip[mysql,postgres]\'\n\nUsage\n-----\n\nThe package installs the `logiskip` command, which takes the following options::\n\n  --source TEXT        URI of source database\n  --destination TEXT   URI of destination database\n  --load-name TEXT     Name of load plugin for migrated application\n  --load-version TEXT  Version of migrated application/schema\n  --dry-run            Roll back transaction instead of commiting\n  -v, --verbosity LVL  Either CRITICAL, ERROR, WARNING, INFO or DEBUG.\n  --help               Show this message and exit.\n\nThe following example migrates an installation of `Roundcube`_ 1.4.1 from MySQL\nto PostgreSQL::\n\n  logiskip --source \'mysql://roundcube:secret@localhost/roundcube\' \\\n           --destination \'postgresql://roundcube:secret@localhost/roundcube\' \\\n\t   --load-name roundcube --load-version 1.4.1\n\nLoads\n-----\n\nLoads in logiskip define migrations for a single application within a version\nconstraint. Here is an artificial example:\n\n.. code-block:: python\n\n   from logiskip.load import BaseLoad\n   from sqlalchemy.orm import sessionmaker\n\n   class ExampleLoad(BaseLoad, name="example", version_constraint=">=1.0,<2.0"):\n       """Load for the application example within the semver contraint 1.x"""\n\n       # Table map for all migrations\n       # Tables mapped to None are skipped\n       default_tables = {\n           "cache": None  # Do not migrate the cache table\n       }\n\n       # Table map for migrating from MySQL to PostgreSQL\n       postgresql_mysql_tables = {\n           "geolocations": None,  # Application supports GIS only in PostgreSQL\n\t   "user": "users"  # Historic naming issue\n       }\n\n       # Field map for the "user" table when migrating from PostgreSQL to MySQL\n       postgresql_mysql_fields_user = {\n           "geolocation_fk": None  # See above\n       }\n\n       def mysql_postgresql_row_users(self, src_table, src_dict):\n           """Do reverse-geolocation for user addresses when migrating to PostgreSQL"""\n\t   # First, do the default conversion\n\t   dest_row = super()._convert_row_default(src_table, src_dict)\n\n\t   # Get geolocation for address\n\t   lat, lon = geocoder.reverse(dest_row["address"]\n\n\t   # Use SQLAlchemy to create a new geolocation entry\n\t   session = sessionmaker(bind=self.dest_engine)()\n\t   geoloc = self.dest_base.classes.geolocations(lat=lat, lon=lon)\n\t   session.commit()\n\n\t   # Set foreign key to geolocation\n\t   dest_row["geolocation_fk"] = geoloc.id\n\t   return dest_row\n\n       # More examples include:\n       #   x_y_field_tablename__fieldname(self, src_value) - Do a conversion on a single field value\n       #   x_y_table_tablename - Do the full table conversion manually\n       # x_y can be default in all places to be used for any migration pair\n\n\nCredits\n-------\n\nlogiskip was sponsored by:\n\n* `credativ GmbH`_\n* `Beuth Hochschule für Technik Berlin`_\n\n::\n\n   Copyright 2021 Dominik George <dominik.george@credativ.de>\n\n   Licensed under the Apache License, Version 2.0 (the "License");\n   you may not use this file except in compliance with the License.\n   You may obtain a copy of the License at\n\n       http://www.apache.org/licenses/LICENSE-2.0\n\n   Unless required by applicable law or agreed to in writing, software\n   distributed under the License is distributed on an "AS IS" BASIS,\n   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.\n   See the License for the specific language governing permissions and\n   limitations under the License.\n\n.. _SQLAlchemy: https://sqlalchemy.org/\n.. _Roundcube: https://roundcube.net/\n.. _credativ GmbH: https://credativ.de/\n.. _Beuth Hochschule für Technik Berlin: https://www.beuth-hochschule.de/\n',
    'author': 'Dominik George',
    'author_email': 'dominik.george@credativ.de',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/credativ/logiskip',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
